# -*- coding: utf-8 -*-
"""
FIXME:
    This class is very old, convoluted, and coupled.
    It really needs to be rewritten efficiently.
    the __setattr__ __getattr__ stuff needs to be redone, and
    DynStruct probably needs to go away.
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import six
from six.moves import cPickle as pickle
try:
    import numpy as np
except ImportError:
    pass
from utool import DynamicStruct
from utool import util_dbg
from utool import util_arg
from utool import util_type
from utool import util_inject
# print, rrr, profile = util_inject.inject(__name__, '[pref]')
util_inject.noinject(__name__, '[pref]')

# ---
# GLOBALS
# ---
PrefNode = DynamicStruct.DynStruct


VERBOSE_PREF = util_arg.get_argflag('--verbpref')


# ---
# Classes
# ---
class PrefInternal(DynamicStruct.DynStruct):
    def __init__(_intern, name, doc, default, hidden, fpath, depeq, choices):
        super(PrefInternal, _intern).__init__(child_exclude_list=[])
        # self._intern describes this node
        _intern.name    = name     # A node has a name
        _intern.doc     = doc      # A node has a name
        _intern.value   = default  # A node has a value
        _intern.hidden  = hidden   # A node can be hidden
        _intern.fpath   = fpath    # A node is cached to
        _intern.depeq   = depeq    # A node depends on
        _intern.editable = True    # A node can be uneditable
        _intern._frozen_type = None     # A node's value type
        # Some preferences are constrained to a list of choices
        if choices is not None:
            _intern.value = PrefChoice(choices, default)

    def get_type(_intern):
        if _intern._frozen_type is not None:
            return _intern._frozen_type
        else:
            return type(_intern.value)

    def freeze_type(_intern):
        _intern._frozen_type = _intern.get_type()


class PrefTree(DynamicStruct.DynStruct):
    def __init__(_tree, parent):
        super(PrefTree, _tree).__init__(child_exclude_list=[])
        # self._tree describes node's children and the parents
        # relationship to them
        _tree.parent               = parent  # Children have parents
        _tree.hidden_children      = []  # Children can be hidden
        _tree.child_list           = []  # There can be many children
        _tree.child_names          = []  # Each child has a name
        _tree.num_visible_children = 0   # Statistic
        _tree.aschildx             = 0   # This node is the x-th child


class PrefChoice(DynamicStruct.DynStruct):
    def __init__(self, choices, default):
        super(PrefChoice, self).__init__(child_exclude_list=[])
        self.choices = choices
        self.sel = 0
        self.change_val(default)

    def change_val(self, new_val):
        # Try to select by index
        if isinstance(new_val, int):
            self.sel = new_val
        # Try to select by value
        elif isinstance(new_val, six.string_types):
            self.sel = self.choices.index(new_val)
        else:
            raise('Exception: Unknown newval=%r' % new_val)
        if self.sel < 0 or self.sel > len(self.choices):
            raise Exception('self.sel=%r is not in the self.choices=%r '
                            % (self.sel, self.choices))

    def combo_val(self):
        return self.choices[self.sel]

    def get_tuple(self):
        return (self.sel, self.choices)


class Pref(PrefNode):
    """
    Structure for Creating Preferences.
    Caveats: When using a value call with ['valname'] to be safe
    Features:
    * Can be saved and loaded.
    * Can be nested
    * Dynamically add/remove
    """
    def __init__(self,
                 default=PrefNode,  # Default value for a Pref is to be itself
                 doc='empty docs',  # Documentation for a preference
                 hidden=False,      # Is a hidden preference?
                 choices=None,      # A list of choices
                 depeq=None,        # List of tuples representing dependencies
                 fpath='',          # Where to save to
                 name='root',       # Name of this node
                 parent=None):      # Reference to parent Pref
        """
        Creates a pref struct that will save itself to pref_fpath if
        available and have init all members of some dictionary
        """
        super(Pref, self).__init__(child_exclude_list=['_intern', '_tree'])
        # Private internal structure
        self._intern = PrefInternal(name, doc, default, hidden, fpath, depeq,
                                    choices)
        self._tree = PrefTree(parent)
        #if default is PrefNode:
        #    print('----------')
        #    print('new Pref(default=PrefNode)')

    def get_type(self):
        return self._intern.get_type()

    def ensure_attr(self, attr, default):
        if not hasattr(self, attr):
            setattr(self, attr, default)
        return getattr(self, attr)

    # -------------------
    # Attribute Setters
    def toggle(self, key):
        """ Toggles a boolean key """
        val = self[key]
        assert isinstance(val, bool), 'key[%r] = %r is not a bool' % (key, val)
        self.pref_update(key, not val)

    def change_combo_val(self, new_val):
        """
        Checks to see if a selection is a valid index or choice of a combo
        preference
        """
        choice_obj = self._intern.value
        assert isinstance(self._intern.value, PrefChoice), 'must be a choice'
        return choice_obj.get_tuple()

    def __overwrite_child_attr(self, name, attr):
        # FIXME: when setting string preference nodes to lists, it casts
        # the list to a string!
        if VERBOSE_PREF:
            print('[pref.__overwrite_child_attr]: %s.%s = %r' % (self._intern.name, name, attr))
        # get child node to "overwrite"
        row = self._tree.child_names.index(name)
        child = self._tree.child_list[row]
        if isinstance(attr, Pref):
            # Do not break pointers when overwriting a Preference
            if issubclass(attr._intern.value, PrefNode):
                # Main Branch Logic
                for (key, val) in six.iteritems(attr):
                    child.__setattr__(key, val)
            else:
                self.__overwrite_child_attr(name, attr.value())
        else:  # Main Leaf Logic:
            #assert(not issubclass(child._intern.type, PrefNode), #(self.full_name() + ' Must be a leaf'))
            # Keep user-readonly map up to date with internals
            if isinstance(child._intern.value, PrefChoice):
                child.change_combo_val(attr)
            else:
                child_type = child._intern.get_type()
                if isinstance(attr, six.string_types) and issubclass(child_type, six.string_types):
                    #import utool as ut
                    #ut.embed()
                    attr = child_type(attr)
                attr_type  = type(attr)
                if attr is not None and child_type is not attr_type:
                    if util_arg.VERBOSE:
                        print('[pref] WARNING TYPE DIFFERENCE!')
                        print('[pref] * expected child_type = %r' % (child_type,))
                        print('[pref] * got attr_type = %r' % (attr_type,))
                        print('[pref] * name = %r' % (name,))
                        print('[pref] * attr = %r' % (attr,))
                    attr = util_type.try_cast(attr, child_type, attr)
                child._intern.value = attr
            self.__dict__[name] = child.value()

    def __new_attr(self, name, attr):
        """
        On a new child attribute:
            1) Check to see if it is wrapped by a Pref object
            2) If not do so, if so add it to the tree structure
        """
        if isinstance(attr, Pref):
            # Child attribute already has a Pref wrapping
            if VERBOSE_PREF:
                print('[pref.__new_attr]: %s.%s = %r' % (self._intern.name, name, attr.value()))
            new_childx = len(self._tree.child_names)
            # Children know about parents
            attr._tree.parent = self     # Give child parent
            attr._intern.name = name     # Give child name
            if attr._intern.depeq is None:
                attr._intern.depeq = self._intern.depeq  # Give child parent dependencies
            if attr._intern.hidden:
                self._tree.hidden_children.append(new_childx)
                self._tree.hidden_children.sort()
            # Used for QTIndexing
            attr._intern.aschildx = new_childx
            # Parents know about children
            self._tree.child_names.append(name)  # Add child to tree
            self._tree.child_list.append(attr)
            self.__dict__[name] = attr.value()   # Add child value to dict
        else:
            # The child attribute is not wrapped. Wrap with Pref and readd.
            pref_attr = Pref(default=attr)
            self.__new_attr(name, pref_attr)

    # Attributes are children
    def __setattr__(self, name, attr):
        """
        Called when an attribute assignment is attempted. This is called instead
        of the normal mechanism (i.e. store the value in the instance
        dictionary). name is the attribute name, value is the value to be
        assigned to it.

        If __setattr__() wants to assign to an instance attribute, it should not
        simply execute self.name = value  this would cause a recursive call to
        itself.  Instead, it should insert the value in the dictionary of
        instance attributes, e.g., self.__dict__[name] = value. For new-style
        classes, rather than accessing the instance dictionary, it should call
        the base class method with the same name, for example,
        object.__setattr__(self, name, value).
        'Wraps child attributes in a Pref object if not already done'
        """
        # No wrapping for private vars: _printable_exclude, _intern, _tree
        if name.find('_') == 0:
            return super(DynamicStruct.DynStruct, self).__setattr__(name, attr)
        # Overwrite if child exists
        if name in self._tree.child_names:
            self.__overwrite_child_attr(name, attr)
        else:
            self.__new_attr(name, attr)

    # -------------------
    # Attribute Getters
    def value(self):
        # Return the wrapper in all its glory
        if self._intern.value == PrefNode:
            return self
        # Return basic types
        elif isinstance(self._intern.value, PrefChoice):
            return self._intern.value.combo_val()
        else:
            return self._intern.value  # TODO AS REFERENCE

    def __getattr__(self, name):
        """
        Called when an attribute lookup has not found the attribute in the usual
        places
        (i.e. it is not an instance attribute nor is it found in the class tree
        for self).
        name is the attribute name.
        This method should return the (computed) attribute value or raise an
        AttributeError exception.
        Get a child from this parent called as last resort.
        Allows easy access to internal prefs
        """
        if name.find('_') == 0:
            # Names that start with underscore actually belong to the Pref object
            return super(PrefNode, self).__getitem__[name]
        if len(name) > 9 and name[-9:] == '_internal':
            # internal names belong to the internal structure
            attrx = self._tree.child_names.index(name[:-9])
            return self._tree.child_list[attrx]
        #print(self._internal.name)
        #print(self._tree)
        try:
            if six.PY2:
                return super(PrefNode, self).__getitem__[name]
            else:
                try:
                    base_self1 = super(PrefNode, self)
                    #base_self2 = super(DynStruct, self)
                    #base_self3 = super()
                    #import utool
                    #utool.embed()
                    return base_self1[name]
                except Exception as ex:
                    print(ex)
                    print('base_self1 = %r' % (base_self1,))
                    #print('base_self2 = %r' % (base_self2,))
                    #print('base_self3 = %r' % (base_self3,))
                    print('name = %r' % (name,))
                    raise
        except Exception as ex:
            if name == 'trait_names':
                # HACK FOR IPYTHON
                pass
            else:
                import utool as ut
                if ut.DEBUG2 and ut.VERBOSE:
                    ut.printex(ex, 'Pref object missing named attribute',
                                  keys=['self._intern.name', 'name'],
                                  iswarning=True)
                raise AttributeError(
                    ('Pref object is missing named attribute: name=%r.'
                     'You might try running ibeis with --nocache-pref '
                     'to see if that fixes things.')  % name)
                #raise

    def iteritems(self):
        """
        Wow this class is messed up. I had to overwrite items when
        moving to python3, just because I haden't called it yet
        """
        for (key, val) in six.iteritems(self.__dict__):
            if key in self._printable_exclude:
                continue
            yield (key, val)

    def items(self):
        """
        Wow this class is messed up. I had to overwrite items when
        moving to python3, just because I haden't called it yet
        """
        for (key, val) in six.iteritems(self.__dict__):
            if key in self._printable_exclude:
                continue
            yield (key, val)

    #----------------
    # Disk caching
    def to_dict(self, split_structs_bit=False):
        """ Converts prefeters to a dictionary.
        Children Pref can be optionally separated """
        pref_dict = {}
        struct_dict = {}
        for (key, val) in six.iteritems(self):
            if split_structs_bit and isinstance(val, Pref):
                struct_dict[key] = val
                continue
            pref_dict[key] = val
        if split_structs_bit:
            return (pref_dict, struct_dict)
        return pref_dict

    asdict = to_dict

    def save(self):
        """ Saves prefs to disk in dict format """
        fpath = self.get_fpath()
        if fpath in ['', None]:
            if self._tree.parent is not None:
                if VERBOSE_PREF:
                    print('[pref.save] Can my parent save me?')  # ...to disk
                return self._tree.parent.save()
            if VERBOSE_PREF:
                print('[pref.save] I cannot be saved. I have no parents.')
            return False
        with open(fpath, 'wb') as f:
            print('[pref] Saving to ' + fpath)
            pref_dict = self.to_dict()
            pickle.dump(pref_dict, f, protocol=2)  # Use protocol 2 to support python2 and 3
        return True

    def get_fpath(self):
        return self._intern.fpath

    def load(self):
        """ Read pref dict stored on disk. Overwriting current values. """
        if VERBOSE_PREF:
            print('[pref.load()]')
            #if not os.path.exists(self._intern.fpath):
            #    msg = '[pref] fpath=%r does not exist' % (self._intern.fpath)
            #    return msg
        fpath = self.get_fpath()
        try:
            with open(fpath, 'rb') as f:
                if VERBOSE_PREF:
                    print('load: %r' % fpath)
                pref_dict = pickle.load(f)
        except EOFError as ex1:
            util_dbg.printex(ex1, 'did not load pref fpath=%r correctly' % fpath, iswarning=True)
            #warnings.warn(msg)
            raise
            #return msg
        except ImportError as ex2:
            util_dbg.printex(ex2, 'did not load pref fpath=%r correctly' % fpath, iswarning=True)
            #warnings.warn(msg)
            raise
            #return msg
        if not util_type.is_dict(pref_dict):
            raise Exception('Preference file is corrupted')
        self.add_dict(pref_dict)
        return True

    #----------------------
    # String representation
    def __str__(self):
        if self._intern.value != PrefNode:
            ret = super(PrefNode, self).__str__()
            #.replace('\n    ', '')
            ret += '\nLEAF ' + repr(self._intern.name) + ':' + repr(self._intern.value)
            return ret
        else:
            ret = repr(self._intern.value)
            return ret

    def full_name(self):
        """ returns name all the way up the tree """
        if self._tree.parent is None:
            return self._intern.name
        return self._tree.parent.full_name() + '.' + self._intern.name

    def get_printable(self, type_bit=True, print_exclude_aug=[]):
        # Remove unsatisfied dependencies from the printed structure
        further_aug = print_exclude_aug[:]
        for child_name in self._tree.child_names:
            depeq = self[child_name + '_internal']._intern.depeq
            if depeq is not None and depeq[0].value() != depeq[1]:
                further_aug.append(child_name)
        return super(Pref, self).get_printable(type_bit, print_exclude_aug=further_aug)

    def customPrintableType(self, name):
        if name in self._tree.child_names:
            row = self._tree.child_names.index(name)
            #child = self._tree.child_list[row]  # child node to "overwrite"
            _typestr = type(self._tree.child_list[row]._intern.value)
            if util_type.is_str(_typestr):
                return _typestr

    def pref_update(self, key, new_val):
        """ Changes a preference value and saves it to disk """
        print('Update and save pref from: %s=%r, to: %s=%r' %
              (key, six.text_type(self[key]), key, six.text_type(new_val)))
        self.__setattr__(key, new_val)
        return self.save()

    def update(self, **kwargs):
        #print('Updating Preference: kwargs = %r' % (kwargs))
        self_keys = set(self.__dict__.keys())
        for key, val in six.iteritems(kwargs):
            if key in self_keys:
                self.__setattr__(key, val)

    # Method for QTWidget
    def createQWidget(self):
        # moving gui code away from utool
        try:
            #from utool._internal.PreferenceWidget import EditPrefWidget
            try:
                from guitool_ibeis.PreferenceWidget import EditPrefWidget
            except ImportError:
                from guitool.PreferenceWidget import EditPrefWidget
            editpref_widget = EditPrefWidget(self)
            editpref_widget.show()
            return editpref_widget
        except ImportError as ex:
            util_dbg.printex(ex, 'Cannot create preference widget. Is guitool and PyQt 4/5 Installed')
            raise

    def qt_get_parent(self):
        return self._tree.parent

    def qt_parents_index_of_me(self):
        return self._tree.aschildx

    def qt_get_child(self, row):
        row_offset = (np.array(self._tree.hidden_children) <= row).sum()
        return self._tree.child_list[row + row_offset]

    def qt_row_count(self):
        return len(self._tree.child_list) - len(self._tree.hidden_children)

    def qt_col_count(self):
        return 2

    def qt_get_data(self, column):
        if column == 0:
            return self._intern.name
        data = self.value()
        if isinstance(data, Pref):  # Recursive Case: Pref
            data = ''
        elif data is None:
            # Check for a get of None
            data = 'None'
        return data

    def qt_is_editable(self):
        uneditable_hack = ['feat_type']
        if self._intern.name in uneditable_hack:
            return False
        if self._intern.depeq is not None:
            return self._intern.depeq[0].value() == self._intern.depeq[1]
        #return self._intern.value is not None
        return self._intern.value != PrefNode

    def qt_set_leaf_data(self, qvar):
        return _qt_set_leaf_data(self, qvar)


def _qt_set_leaf_data(self, qvar):
    """ Sets backend data using QVariants """
    if VERBOSE_PREF:
        print('')
        print('+--- [pref.qt_set_leaf_data]')
        print('[pref.qt_set_leaf_data] qvar = %r' % qvar)
        print('[pref.qt_set_leaf_data] _intern.name=%r' % self._intern.name)
        print('[pref.qt_set_leaf_data] _intern.type_=%r' % self._intern.get_type())
        print('[pref.qt_set_leaf_data] type(_intern.value)=%r' % type(self._intern.value))
        print('[pref.qt_set_leaf_data] _intern.value=%r' % self._intern.value)
        #print('[pref.qt_set_leaf_data] qvar.toString()=%s' % six.text_type(qvar.toString()))
    if self._tree.parent is None:
        raise Exception('[Pref.qtleaf] Cannot set root preference')
    if self.qt_is_editable():
        new_val = '[Pref.qtleaf] BadThingsHappenedInPref'
        if self._intern.value == PrefNode:
            raise Exception('[Pref.qtleaf] Qt can only change leafs')
        elif self._intern.value is None:
            # None could be a number of types
            def cast_order(var, order=[bool, int, float, six.text_type]):
                for type_ in order:
                    try:
                        ret = type_(var)
                        return ret
                    except Exception:
                        continue
            new_val = cast_order(six.text_type(qvar))
        self._intern.get_type()
        if isinstance(self._intern.value, bool):
            #new_val = bool(qvar.toBool())
            print('qvar = %r' % (qvar,))
            new_val = util_type.smart_cast(qvar, bool)
            #new_val = bool(eval(qvar, {}, {}))
            print('new_val = %r' % (new_val,))
        elif isinstance(self._intern.value, int):
            #new_val = int(qvar.toInt()[0])
            new_val = int(qvar)
        # elif isinstance(self._intern.value, float):
        elif self._intern.get_type() in util_type.VALID_FLOAT_TYPES:
            #new_val = float(qvar.toDouble()[0])
            new_val = float(qvar)
        elif isinstance(self._intern.value, six.string_types):
            #new_val = six.text_type(qvar.toString())
            new_val = six.text_type(qvar)
        elif isinstance(self._intern.value, PrefChoice):
            #new_val = qvar.toString()
            new_val = six.text_type(qvar)
            if new_val.upper() == 'NONE':
                new_val = None
        else:
            try:
                #new_val = six.text_type(qvar.toString())
                type_ = self._intern.get_type()
                if type_ is not None:
                    new_val = type_(six.text_type(qvar))
                else:
                    new_val = six.text_type(qvar)
            except Exception:
                raise NotImplementedError(
                    ('[Pref.qtleaf] Unknown internal type. '
                     'type(_intern.value) = %r, '
                     '_intern.get_type() = %r, ')
                    % type(self._intern.value), self._intern.get_type())
        # Check for a set of None
        if isinstance(new_val, six.string_types):
            if new_val.lower() == 'none':
                new_val = None
            elif new_val.lower() == 'true':
                new_val = True
            elif new_val.lower() == 'false':
                new_val = False
        # save to disk after modifying data
        if VERBOSE_PREF:
            print('---')
            print('[pref.qt_set_leaf_data] new_val=%r' % new_val)
            print('[pref.qt_set_leaf_data] type(new_val)=%r' % type(new_val))
            print('L____ [pref.qt_set_leaf_data]')
        # TODO Add ability to set a callback function when certain
        # preferences are changed.
        return self._tree.parent.pref_update(self._intern.name, new_val)
    return 'PrefNotEditable'


def test_Preferences():
    r"""
    CommandLine:
        python -m utool.Preferences --test-test_Preferences --show --verbpref

    Example:
        >>> # DISABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(module:guitool_ibeis)
        >>> # FIXME depends on guitool_ibei
        >>> from utool.Preferences import *  # NOQA
        >>> import utool as ut
        >>> import guitool_ibeis
        >>> guitool_ibeis.ensure_qtapp()
        >>> root = test_Preferences()
        >>> ut.quit_if_noshow()
        >>> widget = root.createQWidget()
        >>> #widget.show()
        >>> guitool_ibeis.qtapp_loop(widget)
    """
    root = Pref()
    root.a = Pref()
    root.strvar = 'foobar'
    root.intvar = 1
    root.floatvar = -1.3
    root.boolvar = True
    root.listvar = [True, 1, 3]

    root.a.strvar = 'foobar1'
    root.a.intvar = -1
    root.a.floatvar = 2.4
    root.a.boolvar2 = False
    return root


if __name__ == '__main__':
    """
    CommandLine:
        python -m utool.Preferences
        python -m utool.Preferences --allexamples
        python -m utool.Preferences --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""

flake8: noqa

script to help me figure out where things are used.
it is not named well

set PATH=%HOME%\code\utool\utool\util_scripts;%PATH%
classfuncs.py %HOME%/code/ibeis/ibeis/control/IBEISControl.py %HOME%/code/ibeis

classfuncs.py C:/Users/joncrall/code/ibeis/ibeis/algo/hots/query_request.py %HOME%/code/ibeis
"""
from __future__ import absolute_import, division, print_function
import utool as ut
import sys


def show_function_usage(fname, funcname_list, dpath_list):
    # Check to see for function usage
    funcname_list = [r'\b%s\b' % (funcname.strip(),) for funcname in funcname_list if len(funcname) > 0]
    flagged_funcnames = []
    for funcname in funcname_list:
        found_filestr_list, found_lines_list, found_lxs_list = ut.grep([funcname], dpath_list=dpath_list)
        total = 0
        for lines in found_lines_list:
            total += len(lines)
        funcname_ = funcname.replace('\\b', '')
        print(funcname_ + ' ' + str(total))
        if total == 1:
            flagged_funcnames.append(funcname_)
        # See where external usage is
        isexternal_list = [fname == fname_ for fname_ in found_filestr_list]
        external_filestr_list = ut.compress(found_filestr_list, isexternal_list)
        external_lines_list = ut.compress(found_lines_list, isexternal_list)
        #external_lxs_list = ut.compress(found_lxs_list, isexternal_list)
        if len(external_filestr_list) == 0:
            print(' no external usage')
        else:
            for filename, lines in zip(external_filestr_list, external_lines_list):
                print(' * filename=%r' % (filename,))
                print(ut.repr4(lines))
            #print(ut.repr4(list(zip(external_filestr_list, external_lines_list))))
    print('----------')
    print('flagged:')
    print('\n'.join(flagged_funcnames))
    #ut.embed()


def get_codedir():
    from os.path import dirname, realpath
    import os
    if 'CODE_DIR' in os.environ:
        CODE_DIR = os.environ.get('CODE_DIR')
    else:
        CODE_DIR = dirname(dirname(realpath(__file__)))   # Home is where the .. is.  # '~/code'
    return CODE_DIR


if __name__ == '__main__':

    CODE_DIR = get_codedir()
    rman = ut.RepoManager(repo_urls=[
        'https://github.com/Erotemic/utool.git',
        'https://github.com/Erotemic/guitool.git',
        'https://github.com/Erotemic/plottool.git',
        'https://github.com/Erotemic/vtool.git',
        'https://github.com/bluemellophone/detecttools.git',
        'https://github.com/Erotemic/hesaff.git',
        'https://github.com/bluemellophone/pyrf.git',
        'https://github.com/Erotemic/ibeis.git',
        'https://github.com/aweinstock314/cyth.git',
        #'https://github.com/hjweide/pygist',
    ], code_dir=CODE_DIR)

    # (IBEIS_REPO_URLS, IBEIS_REPO_DIRS) = ut.repo_list(, forcessh=False)
    # ut.set_project_repos(IBEIS_REPO_URLS, IBEIS_REPO_DIRS)
    dpath_list = rman.repo_dirs
    # IBEIS_REPO_DIRS

    fname = ut.truepath(sys.argv[1])
    #if len(sys.argv) >= 3:
    #    grep_dpath = ut.truepath(sys.argv[2])
    print('Classfuncs of %r' % fname)
    funcname_list = ut.list_class_funcnames(fname)
    funcname_list = ut.list_global_funcnames(fname)
    print(ut.indentjoin(funcname_list, '\n *   '))
    show_function_usage(fname, funcname_list, dpath_list)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
import sys
try:
    import numpy as np
except ImportError as ex:
    pass
import decimal
from utool import util_type
from utool import util_inject
print, rrr, profile = util_inject.inject2(__name__)


def order_of_magnitude_ceil(num):
    nDigits = np.ceil(np.log10(num))
    scalefactor = 10 ** (nDigits - 1)
    return np.ceil(num / scalefactor) * scalefactor


def float_to_decimal(f):
    """
    Convert a floating point number to a Decimal with no loss of information

    References:
        http://docs.python.org/library/decimal.html#decimal-faq
    """
    n, d = f.as_integer_ratio()
    numerator, denominator = decimal.Decimal(n), decimal.Decimal(d)
    ctx = decimal.Context(prec=60)
    result = ctx.divide(numerator, denominator)
    while ctx.flags[decimal.Inexact]:
        ctx.flags[decimal.Inexact] = False
        ctx.prec *= 2
        result = ctx.divide(numerator, denominator)
    return result


def sigfig_str(number, sigfig):
    """
    References:
        http://stackoverflow.com/questions/2663612/nicely-repr-float-in-python
    """
    assert(sigfig > 0)
    try:
        d = decimal.Decimal(number)
    except TypeError:
        d = float_to_decimal(float(number))
    sign, digits, exponent = d.as_tuple()
    if len(digits) < sigfig:
        digits = list(digits)
        digits.extend([0] * (sigfig - len(digits)))
    shift = d.adjusted()
    result = int(''.join(map(str, digits[:sigfig])))
    # Round the result
    if len(digits) > sigfig and digits[sigfig] >= 5:
        result += 1
    result = list(str(result))
    # Rounding can change the length of result
    # If so, adjust shift
    shift += len(result) - sigfig
    # reset len of result to sigfig
    result = result[:sigfig]
    if shift >= sigfig - 1:
        # Tack more zeros on the end
        result += ['0'] * (shift - sigfig + 1)
    elif 0 <= shift:
        # Place the decimal point in between digits
        result.insert(shift + 1, '.')
    else:
        # Tack zeros on the front
        assert(shift < 0)
        result = ['0.'] + ['0'] * (-shift - 1) + result
    if sign:
        result.insert(0, '-')
    return ''.join(result)


def num2_sigfig(num):
    return int(np.ceil(np.log10(num)))


def num_fmt(num, max_digits=None):
    r"""
    Weird function. Not very well written. Very special case-y

    Args:
        num (int or float):
        max_digits (int):

    Returns:
        str:

    CommandLine:
        python -m utool.util_num --test-num_fmt

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_num import *  # NOQA
        >>> # build test data
        >>> num_list = [0, 0.0, 1.2, 1003232, 41431232., .0000000343, -.443243]
        >>> max_digits = None
        >>> # execute function
        >>> result = [num_fmt(num, max_digits) for num in num_list]
        >>> # verify results
        >>> print(result)
        ['0', '0.0', '1.2', '1,003,232', '41431232.0', '0.0', '-0.443']
    """
    if num is None:
        return 'None'
    def num_in_mag(num, mag):
        return mag > num and num > (-1 * mag)
    if max_digits is None:
        # TODO: generalize
        if num_in_mag(num, 1):
            if num_in_mag(num, .1):
                max_digits = 4
            else:
                max_digits = 3
        else:
            max_digits = 1
    if util_type.is_float(num):
        num_str = ('%.' + str(max_digits) + 'f') % num
        # Handle trailing and leading zeros
        num_str = num_str.rstrip('0').lstrip('0')
        if num_str.startswith('.'):
            num_str = '0' + num_str
        if num_str.endswith('.'):
            num_str = num_str + '0'
        return num_str
    elif util_type.is_int(num):
        return int_comma_str(num)
    else:
        return '%r'


def int_comma_str(num):
    int_str = ''
    reversed_digits = decimal.Decimal(num).as_tuple()[1][::-1]
    for i, digit in enumerate(reversed_digits):
        if (i) % 3 == 0 and i != 0:
            int_str += ','
        int_str += str(digit)
    return int_str[::-1]


def get_sys_maxint():
    return sys.maxint


def get_sys_minint():
    return sys.maxint + 1


def get_sys_maxfloat():
    return sys.float_info.max



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import six
global_cache_fname = 'global_cache.shelf'
global_cache_dname = 'global_cache' + ('' if six.PY2 else '_py3')
default_appname = 'utool'



# -*- coding: utf-8 -*-
# TODO Licence:
#
# TODO:  move library intensive functions to vtool
from __future__ import absolute_import, division, print_function, unicode_literals
# import operator
import six
# from six.moves import zip, range, reduce
# from utool import util_type
from utool import util_inject
# from utool import util_decor
try:
    import numpy as np
    HAVE_NUMPY = True
except ImportError:
    HAVE_NUMPY = False
    # TODO remove numpy
    pass
try:
    import scipy.spatial.distance as spdist  # NOQA
    HAVE_SCIPY = True
except ImportError:
    HAVE_SCIPY = False
print, rrr, profile = util_inject.inject2(__name__)


unicode = six.text_type

PHI = 1.61803398875
PHI_A = (1 / PHI)
PHI_B = 1 - PHI_A


def bayesnet():
    """
    References:
        https://class.coursera.org/pgm-003/lecture/17
        http://www.cs.ubc.ca/~murphyk/Bayes/bnintro.html
        http://www3.cs.stonybrook.edu/~sael/teaching/cse537/Slides/chapter14d_BP.pdf
        http://www.cse.unsw.edu.au/~cs9417ml/Bayes/Pages/PearlPropagation.html
        https://github.com/pgmpy/pgmpy.git
        http://pgmpy.readthedocs.org/en/latest/
        http://nipy.bic.berkeley.edu:5000/download/11
    """
    import utool as ut
    # import operator as op
    # # Enumerate all possible events
    # varcard_list = list(map(op.attrgetter('variable_card'), cpd_list))
    # _esdat = list(ut.iprod(*map(range, varcard_list)))
    # _escol = list(map(op.attrgetter('variable'), cpd_list))
    # event_space = pd.DataFrame(_esdat, columns=_escol)

    # # Custom compression of event space to inspect a specific graph
    # def compress_space_flags(event_space, var1, var2, var3, cmp12_):
    #     """
    #     var1, var2, cmp_ = 'Lj', 'Lk', op.eq
    #     """
    #     import vtool as vt
    #     data = event_space
    #     other_cols = ut.setdiff_ordered(data.columns.tolist(), [var1, var2, var3])
    #     case_flags12 = cmp12_(data[var1], data[var2]).values
    #     # case_flags23 = cmp23_(data[var2], data[var3]).values
    #     # case_flags = np.logical_and(case_flags12, case_flags23)
    #     case_flags = case_flags12
    #     case_flags = case_flags.astype(np.int64)
    #     subspace = np.hstack((case_flags[:, None], data[other_cols].values))
    #     sel_ = vt.unique_row_indexes(subspace)
    #     flags = np.logical_and(mask, case_flags)
    #     return flags

    # # Build special cases
    # case_same   = event_space.loc[compress_space_flags(event_space, 'Li', 'Lj', 'Lk', op.eq)]
    # case_diff = event_space.loc[compress_space_flags(event_space, 'Li', 'Lj', 'Lk', op.ne)]
    # special_cases = [
    #     case_same,
    #     case_diff,
    # ]

    from pgmpy.factors import TabularCPD
    from pgmpy.models import BayesianModel
    import pandas as pd
    from pgmpy.inference import BeliefPropagation  # NOQA
    from pgmpy.inference import VariableElimination  # NOQA

    name_nice = ['n1', 'n2', 'n3']
    score_nice = ['low', 'high']
    match_nice = ['diff', 'same']
    num_names = len(name_nice)
    num_scores = len(score_nice)
    nid_basis = list(range(num_names))
    score_basis = list(range(num_scores))

    semtype2_nice = {
        'score': score_nice,
        'name': name_nice,
        'match': match_nice,
    }
    var2_cpd = {
    }
    globals()['semtype2_nice'] = semtype2_nice
    globals()['var2_cpd'] = var2_cpd

    name_combo = np.array(list(ut.iprod(nid_basis, nid_basis)))
    combo_is_same = name_combo.T[0] == name_combo.T[1]
    def get_expected_scores_prob(level1, level2):
        part1 = combo_is_same * level1
        part2 = (1 - combo_is_same) * (1 - (level2))
        expected_scores_level = part1 + part2
        return expected_scores_level

    # def make_cpd():

    def name_cpd(aid):
        from pgmpy.factors import TabularCPD
        cpd = TabularCPD(
            variable='N' + aid,
            variable_card=num_names,
            values=[[1.0 / num_names] * num_names])
        cpd.semtype = 'name'
        return cpd

    name_cpds = [name_cpd('i'), name_cpd('j'), name_cpd('k')]
    var2_cpd.update(dict(zip([cpd.variable for cpd in name_cpds], name_cpds)))
    if True:
        num_same_diff = 2
        samediff_measure = np.array([
            # get_expected_scores_prob(.12, .2),
            # get_expected_scores_prob(.88, .8),
            get_expected_scores_prob(0, 0),
            get_expected_scores_prob(1, 1),
        ])
        samediff_vals = (samediff_measure / samediff_measure.sum(axis=0)).tolist()
        def samediff_cpd(aid1, aid2):
            cpd = TabularCPD(
                variable='A' + aid1 + aid2,
                variable_card=num_same_diff,
                values=samediff_vals,
                evidence=['N' + aid1, 'N' + aid2],  # [::-1],
                evidence_card=[num_names, num_names])  # [::-1])
            cpd.semtype = 'match'
            return cpd
        samediff_cpds = [samediff_cpd('i', 'j'), samediff_cpd('j', 'k'), samediff_cpd('k', 'i')]
        var2_cpd.update(dict(zip([cpd.variable for cpd in samediff_cpds], samediff_cpds)))

        if True:
            def score_cpd(aid1, aid2):
                semtype = 'score'
                evidence = ['A' + aid1 + aid2, 'N' + aid1, 'N' + aid2]
                evidence_cpds = [var2_cpd[key] for key in evidence]
                evidence_nice = [semtype2_nice[cpd.semtype] for cpd in evidence_cpds]
                evidence_card = list(map(len, evidence_nice))
                evidence_states = list(ut.iprod(*evidence_nice))
                variable_basis = semtype2_nice[semtype]

                variable_values = []
                for mystate in variable_basis:
                    row = []
                    for state in evidence_states:
                        if state[0] == state[1]:
                            if state[2] == 'same':
                                val = .2 if mystate == 'low' else .8
                            else:
                                val = 1
                                # val = .5 if mystate == 'low' else .5
                        elif state[0] != state[1]:
                            if state[2] == 'same':
                                val = .5 if mystate == 'low' else .5
                            else:
                                val = 1
                                # val = .9 if mystate == 'low' else .1
                        row.append(val)
                    variable_values.append(row)

                cpd = TabularCPD(
                    variable='S' + aid1 + aid2,
                    variable_card=len(variable_basis),
                    values=variable_values,
                    evidence=evidence,  # [::-1],
                    evidence_card=evidence_card)  # [::-1])
                cpd.semtype = semtype
                return cpd
        else:
            score_values = [
                [.8, .1],
                [.2, .9],
            ]
            def score_cpd(aid1, aid2):
                cpd = TabularCPD(
                    variable='S' + aid1 + aid2,
                    variable_card=num_scores,
                    values=score_values,
                    evidence=['A' + aid1 + aid2],  # [::-1],
                    evidence_card=[num_same_diff])  # [::-1])
                cpd.semtype = 'score'
                return cpd

        score_cpds = [score_cpd('i', 'j'), score_cpd('j', 'k')]
        cpd_list = name_cpds + score_cpds + samediff_cpds
    else:
        score_measure = np.array([get_expected_scores_prob(level1, level2)
                                  for level1, level2 in
                                  zip(np.linspace(.1, .9, num_scores),
                                      np.linspace(.2, .8, num_scores))])

        score_values = (score_measure / score_measure.sum(axis=0)).tolist()

        def score_cpd(aid1, aid2):
            cpd = TabularCPD(
                variable='S' + aid1 + aid2,
                variable_card=num_scores,
                values=score_values,
                evidence=['N' + aid1, 'N' + aid2],
                evidence_card=[num_names, num_names])
            cpd.semtype = 'score'
            return cpd
        score_cpds = [score_cpd('i', 'j'), score_cpd('j', 'k')]
        cpd_list = name_cpds + score_cpds
        pass

    input_graph = []
    for cpd in cpd_list:
        if cpd.evidence is not None:
            for evar in cpd.evidence:
                input_graph.append((evar, cpd.variable))
    name_model = BayesianModel(input_graph)
    name_model.add_cpds(*cpd_list)

    var2_cpd.update(dict(zip([cpd.variable for cpd in cpd_list], cpd_list)))
    globals()['var2_cpd'] = var2_cpd

    varnames = [cpd.variable for cpd in cpd_list]

    # --- PRINT CPDS ---

    cpd = score_cpds[0]
    def print_cpd(cpd):
        print('CPT: %r' % (cpd,))
        index = semtype2_nice[cpd.semtype]
        if cpd.evidence is None:
            columns = ['None']
        else:
            basis_lists = [semtype2_nice[var2_cpd[ename].semtype] for ename in cpd.evidence]
            columns = [','.join(x) for x in ut.iprod(*basis_lists)]
        data = cpd.get_cpd()
        print(pd.DataFrame(data, index=index, columns=columns))

    for cpd in name_model.get_cpds():
        print('----')
        print(cpd._str('phi'))
        print_cpd(cpd)

    # --- INFERENCE ---

    Ni = name_cpds[0]

    event_space_combos = {}
    event_space_combos[Ni.variable] = 0  # Set ni to always be Fred
    for cpd in cpd_list:
        if cpd.semtype == 'score':
            event_space_combos[cpd.variable] = list(range(cpd.variable_card))
    evidence_dict = ut.all_dict_combinations(event_space_combos)

    # Query about name of annotation k given different event space params

    def pretty_evidence(evidence):
        return [key + '=' + str(semtype2_nice[var2_cpd[key].semtype][val])
                for key, val in evidence.items()]

    def print_factor(factor):
        row_cards = factor.cardinality
        row_vars = factor.variables
        values = factor.values.reshape(np.prod(row_cards), 1).flatten()
        # col_cards = 1
        # col_vars = ['']
        basis_lists = list(zip(*list(ut.iprod(*[range(c) for c in row_cards]))))
        nice_basis_lists = []
        for varname, basis in zip(row_vars, basis_lists):
            cpd = var2_cpd[varname]
            _nice_basis = ut.take(semtype2_nice[cpd.semtype], basis)
            nice_basis = ['%s=%s' % (varname, val) for val in _nice_basis]
            nice_basis_lists.append(nice_basis)
        row_lbls = [', '.join(sorted(x)) for x in zip(*nice_basis_lists)]
        print(ut.repr3(dict(zip(row_lbls, values)), precision=3, align=True, key_order_metric='-val'))

    # name_belief = BeliefPropagation(name_model)
    name_belief = VariableElimination(name_model)
    import pgmpy
    import six  # NOQA

    def try_query(evidence):
        print('--------')
        query_vars = ut.setdiff_ordered(varnames, list(evidence.keys()))
        evidence_str = ', '.join(pretty_evidence(evidence))
        probs = name_belief.query(query_vars, evidence)
        factor_list = probs.values()
        joint_factor = pgmpy.factors.factor_product(*factor_list)
        print('P(' + ', '.join(query_vars) + ' | ' + evidence_str + ')')
        # print(six.text_type(joint_factor))
        factor = joint_factor  # NOQA
        # print_factor(factor)
        # import utool as ut
        print(ut.hz_str([(f._str(phi_or_p='phi')) for f in factor_list]))

    for evidence in evidence_dict:
        try_query(evidence)

    evidence = {'Aij': 1, 'Ajk': 1, 'Aki': 1, 'Ni': 0}
    try_query(evidence)

    evidence = {'Aij': 0, 'Ajk': 0, 'Aki': 0, 'Ni': 0}
    try_query(evidence)

    globals()['score_nice'] = score_nice
    globals()['name_nice'] = name_nice
    globals()['score_basis'] = score_basis
    globals()['nid_basis'] = nid_basis

    print('Independencies')
    print(name_model.get_independencies())
    print(name_model.local_independencies([Ni.variable]))

    # name_belief = BeliefPropagation(name_model)
    # # name_belief = VariableElimination(name_model)
    # for case in special_cases:
    #     test_data = case.drop('Lk', axis=1)
    #     test_data = test_data.reset_index(drop=True)
    #     print('----')
    #     for i in range(test_data.shape[0]):
    #         evidence = test_data.loc[i].to_dict()
    #         probs = name_belief.query(['Lk'], evidence)
    #         factor = probs['Lk']
    #         probs = factor.values
    #         evidence_ = evidence.copy()
    #         evidence_['Li'] = name_nice[evidence['Li']]
    #         evidence_['Lj'] = name_nice[evidence['Lj']]
    #         evidence_['Sij'] = score_nice[evidence['Sij']]
    #         evidence_['Sjk'] = score_nice[evidence['Sjk']]
    #         nice2_prob = ut.odict(zip(name_nice, probs.tolist()))
    #         ut.print_python_code('P(Lk | {evidence}) = {cpt}'.format(
    #             evidence=(ut.repr2(evidence_, explicit=True, nobraces=True, strvals=True)),
    #             cpt=ut.repr3(nice2_prob, precision=3, align=True, key_order_metric='-val')
    #         ))

    # for case in special_cases:
    #     test_data = case.drop('Lk', axis=1)
    #     test_data = test_data.drop('Lj', axis=1)
    #     test_data = test_data.reset_index(drop=True)
    #     print('----')
    #     for i in range(test_data.shape[0]):
    #         evidence = test_data.loc[i].to_dict()
    #         query_vars = ['Lk', 'Lj']
    #         probs = name_belief.query(query_vars, evidence)
    #         for queryvar in query_vars:
    #             factor = probs[queryvar]
    #             print(factor._str('phi'))
    #             probs = factor.values
    #             evidence_ = evidence.copy()
    #             evidence_['Li'] = name_nice[evidence['Li']]
    #             evidence_['Sij'] = score_nice[evidence['Sij']]
    #             evidence_['Sjk'] = score_nice[evidence['Sjk']]
    #             nice2_prob = ut.odict(zip([queryvar + '=' + x for x in name_nice], probs.tolist()))
    #             ut.print_python_code('P({queryvar} | {evidence}) = {cpt}'.format(
    #                 query_var=query_var,
    #                 evidence=(ut.repr2(evidence_, explicit=True, nobraces=True, strvals=True)),
    #                 cpt=ut.repr3(nice2_prob, precision=3, align=True, key_order_metric='-val')
    #             ))

    # _ draw model

    import plottool as pt
    import networkx as netx
    fig = pt.figure()  # NOQA
    fig.clf()
    ax = pt.gca()

    netx_nodes = [(node, {}) for node in name_model.nodes()]
    netx_edges = [(etup[0], etup[1], {}) for etup in name_model.edges()]
    netx_graph = netx.DiGraph()
    netx_graph.add_nodes_from(netx_nodes)
    netx_graph.add_edges_from(netx_edges)

    # pos = netx.graphviz_layout(netx_graph)
    pos = netx.pydot_layout(netx_graph, prog='dot')
    netx.draw(netx_graph, pos=pos, ax=ax, with_labels=True)

    pt.plt.savefig('foo.png')
    ut.startfile('foo.png')


def bayesnet_examples():
    from pgmpy.factors import TabularCPD
    from pgmpy.models import BayesianModel
    import pandas as pd

    student_model = BayesianModel([('D', 'G'),
                                   ('I', 'G'),
                                   ('G', 'L'),
                                   ('I', 'S')])
    # we can generate some random data.
    raw_data = np.random.randint(low=0, high=2, size=(1000, 5))
    data = pd.DataFrame(raw_data, columns=['D', 'I', 'G', 'L', 'S'])
    data_train = data[: int(data.shape[0] * 0.75)]
    student_model.fit(data_train)
    student_model.get_cpds()

    data_test = data[int(0.75 * data.shape[0]): data.shape[0]]
    data_test.drop('D', axis=1, inplace=True)
    student_model.predict(data_test)

    grade_cpd = TabularCPD(
        variable='G',
        variable_card=3,
        values=[[0.3, 0.05, 0.9, 0.5],
                [0.4, 0.25, 0.08, 0.3],
                [0.3, 0.7, 0.02, 0.2]],
        evidence=['I', 'D'],
        evidence_card=[2, 2])
    difficulty_cpd = TabularCPD(
        variable='D',
        variable_card=2,
        values=[[0.6, 0.4]])
    intel_cpd = TabularCPD(
        variable='I',
        variable_card=2,
        values=[[0.7, 0.3]])
    letter_cpd = TabularCPD(
        variable='L',
        variable_card=2,
        values=[[0.1, 0.4, 0.99],
                [0.9, 0.6, 0.01]],
        evidence=['G'],
        evidence_card=[3])
    sat_cpd = TabularCPD(
        variable='S',
        variable_card=2,
        values=[[0.95, 0.2],
                [0.05, 0.8]],
        evidence=['I'],
        evidence_card=[2])
    student_model.add_cpds(grade_cpd, difficulty_cpd,
                           intel_cpd, letter_cpd,
                           sat_cpd)



# -*- coding: utf-8 -*-
"""
Check that SendInput can work the way we want it to

The tips and tricks at http://www.pinvoke.net/default.aspx/user32.sendinput
is useful!

"""
from __future__ import absolute_import, division, print_function, unicode_literals
import time
import ctypes
import six

__all__ = ['KeySequenceError', 'SendKeys']

unicode = six.text_type


#pylint: disable-msg=R0903

DEBUG = 0

MapVirtualKey = ctypes.windll.user32.MapVirtualKeyW
SendInput = ctypes.windll.user32.SendInput
VkKeyScan = ctypes.windll.user32.VkKeyScanW
VkKeyScan.restype = ctypes.c_short
VkKeyScan.argtypes = [ctypes.c_wchar]

DWORD = ctypes.c_ulong
LONG = ctypes.c_long
WORD = ctypes.c_ushort


# C:/PROGRA~1/MICROS~4/VC98/Include/winuser.h 4283
class MOUSEINPUT(ctypes.Structure):
    "Needed for complete definition of INPUT structure - not used"
    _pack_ = 2
    _fields_ = [
        # C:/PROGRA~1/MICROS~4/VC98/Include/winuser.h 4283
        ('dx', LONG),
        ('dy', LONG),
        ('mouseData', DWORD),
        ('dwFlags', DWORD),
        ('time', DWORD),
        ('dwExtraInfo', DWORD),
    ]
assert ctypes.sizeof(MOUSEINPUT) == 24, ctypes.sizeof(MOUSEINPUT)
assert ctypes.alignment(MOUSEINPUT) == 2, ctypes.alignment(MOUSEINPUT)


# C:/PROGRA~1/MICROS~4/VC98/Include/winuser.h 4292
class KEYBDINPUT(ctypes.Structure):
    "A particular keyboard event"
    _pack_ = 2
    _fields_ = [
        # C:/PROGRA~1/MICROS~4/VC98/Include/winuser.h 4292
        ('wVk', WORD),
        ('wScan', WORD),
        ('dwFlags', DWORD),
        ('time', DWORD),
        ('dwExtraInfo', DWORD),
    ]
assert ctypes.sizeof(KEYBDINPUT) == 16, ctypes.sizeof(KEYBDINPUT)
assert ctypes.alignment(KEYBDINPUT) == 2, ctypes.alignment(KEYBDINPUT)


class HARDWAREINPUT(ctypes.Structure):
    "Needed for complete definition of INPUT structure - not used"
    _pack_ = 2
    _fields_ = [
        # C:/PROGRA~1/MICROS~4/VC98/Include/winuser.h 4300
        ('uMsg', DWORD),
        ('wParamL', WORD),
        ('wParamH', WORD),
    ]
assert ctypes.sizeof(HARDWAREINPUT) == 8, ctypes.sizeof(HARDWAREINPUT)
assert ctypes.alignment(HARDWAREINPUT) == 2, ctypes.alignment(HARDWAREINPUT)


# C:/PROGRA~1/MICROS~4/VC98/Include/winuser.h 4314
class UNION_INPUT_STRUCTS(ctypes.Union):
    "The C Union type representing a single Event of any type"
    _fields_ = [
        # C:/PROGRA~1/MICROS~4/VC98/Include/winuser.h 4314
        ('mi', MOUSEINPUT),
        ('ki', KEYBDINPUT),
        ('hi', HARDWAREINPUT),
    ]
assert ctypes.sizeof(UNION_INPUT_STRUCTS) == 24, \
    ctypes.sizeof(UNION_INPUT_STRUCTS)
assert ctypes.alignment(UNION_INPUT_STRUCTS) == 2, \
    ctypes.alignment(UNION_INPUT_STRUCTS)


# C:/PROGRA~1/MICROS~4/VC98/Include/winuser.h 4310
class INPUT(ctypes.Structure):
    "See: http://msdn.microsoft.com/en-us/library/ms646270%28VS.85%29.aspx"
    _pack_ = 2
    _fields_ = [
        # C:/PROGRA~1/MICROS~4/VC98/Include/winuser.h 4310
        ('type', DWORD),
        # Unnamed field renamed to '_'
        ('_', UNION_INPUT_STRUCTS),
    ]
assert ctypes.sizeof(INPUT) == 28, ctypes.sizeof(INPUT)
assert ctypes.alignment(INPUT) == 2, ctypes.alignment(INPUT)


INPUT_KEYBOARD = 1
KEYEVENTF_EXTENDEDKEY = 1
KEYEVENTF_KEYUP       = 2
KEYEVENTF_UNICODE     = 4
KEYEVENTF_SCANCODE    = 8
VK_SHIFT        = 16
VK_CONTROL      = 17
VK_MENU         = 18

# 'codes' recognized as {CODE( repeat)?}
CODES = {
    'BACK':     8,
    'BACKSPACE': 8,
    'BKSP':     8,
    'BREAK':    3,
    'BS':       8,
    'CAP':      20,
    'CAPSLOCK': 20,
    'DEL':      46,
    'DELETE':   46,
    'DOWN':     40,
    'END':      35,
    'ENTER':    13,
    'ESC':      27,
    'F1':       112,
    'F2':       113,
    'F3':       114,
    'F4':       115,
    'F5':       116,
    'F6':       117,
    'F7':       118,
    'F8':       119,
    'F9':       120,
    'F10':      121,
    'F11':      122,
    'F12':      123,
    'F13':      124,
    'F14':      125,
    'F15':      126,
    'F16':      127,
    'F17':      128,
    'F18':      129,
    'F19':      130,
    'F20':      131,
    'F21':      132,
    'F22':      133,
    'F23':      134,
    'F24':      135,
    'HELP':     47,
    'HOME':     36,
    'INS':      45,
    'INSERT':   45,
    'LEFT':     37,
    'LWIN':     91,
    'NUMLOCK':  144,
    'PGDN':     34,
    'PGUP':     33,
    'PRTSC':    44,
    'RIGHT':    39,
    'RMENU':    165,
    'RWIN':     92,
    'SCROLLLOCK': 145,
    'SPACE':     32,
    'TAB':       9,
    'UP':        38,

    'VK_ACCEPT': 30,
    'VK_ADD':    107,
    'VK_APPS':    93,
    'VK_ATTN':    246,
    'VK_BACK':    8,
    'VK_CANCEL':  3,
    'VK_CAPITAL': 20,
    'VK_CLEAR':   12,
    'VK_CONTROL': 17,
    'VK_CONVERT': 28,
    'VK_CRSEL':   247,
    'VK_DECIMAL': 110,
    'VK_DELETE':  46,
    'VK_DIVIDE':  111,
    'VK_DOWN':    40,
    'VK_END':     35,
    'VK_EREOF':   249,
    'VK_ESCAPE':  27,
    'VK_EXECUTE': 43,
    'VK_EXSEL':   248,
    'VK_F1':      112,
    'VK_F2':      113,
    'VK_F3':      114,
    'VK_F4':      115,
    'VK_F5':      116,
    'VK_F6':      117,
    'VK_F7':      118,
    'VK_F8':      119,
    'VK_F9':      120,
    'VK_F10':     121,
    'VK_F11':     122,
    'VK_F12':     123,
    'VK_F13':     124,
    'VK_F14':     125,
    'VK_F15':     126,
    'VK_F16':     127,
    'VK_F17':     128,
    'VK_F18':     129,
    'VK_F19':     130,
    'VK_F20':     131,
    'VK_F21':     132,
    'VK_F22':     133,
    'VK_F23':     134,
    'VK_F24':     135,
    'VK_FINAL':   24,
    'VK_HANGEUL':  21,
    'VK_HANGUL':   21,
    'VK_HANJA':    25,
    'VK_HELP':     47,
    'VK_HOME':     36,
    'VK_INSERT':   45,
    'VK_JUNJA':    23,
    'VK_KANA':     21,
    'VK_KANJI':    25,
    'VK_LBUTTON':   1,
    'VK_LCONTROL': 162,
    'VK_LEFT':     37,
    'VK_LMENU':   164,
    'VK_LSHIFT':  160,
    'VK_LWIN':     91,
    'VK_MBUTTON':    4,
    'VK_MENU':        18,
    'VK_MODECHANGE':  31,
    'VK_MULTIPLY':   106,
    'VK_NEXT':        34,
    'VK_NONAME':     252,
    'VK_NONCONVERT':  29,
    'VK_NUMLOCK':    144,
    'VK_NUMPAD0':     96,
    'VK_NUMPAD1':     97,
    'VK_NUMPAD2':     98,
    'VK_NUMPAD3':     99,
    'VK_NUMPAD4':    100,
    'VK_NUMPAD5':    101,
    'VK_NUMPAD6':    102,
    'VK_NUMPAD7':    103,
    'VK_NUMPAD8':    104,
    'VK_NUMPAD9':    105,
    'VK_OEM_CLEAR':  254,
    'VK_PA1':        253,
    'VK_PAUSE':       19,
    'VK_PLAY':       250,
    'VK_PRINT':       42,
    'VK_PRIOR':       33,
    'VK_PROCESSKEY': 229,
    'VK_RBUTTON':      2,
    'VK_RCONTROL':   163,
    'VK_RETURN':      13,
    'VK_RIGHT':       39,
    'VK_RMENU':      165,
    'VK_RSHIFT':     161,
    'VK_RWIN':        92,
    'VK_SCROLL':     145,
    'VK_SELECT':      41,
    'VK_SEPARATOR':  108,
    'VK_SHIFT':       16,
    'VK_SNAPSHOT':    44,
    'VK_SPACE':       32,
    'VK_SUBTRACT':   109,
    'VK_TAB':          9,
    'VK_UP':          38,
    'ZOOM':          251,
}
# reverse the CODES dict to make it easy to look up a particular code name
CODE_NAMES = dict((entry[1], entry[0]) for entry in CODES.items())

# modifier keys
MODIFIERS = {
    '+': VK_SHIFT,
    '^': VK_CONTROL,
    '%': VK_MENU,
}


class KeySequenceError(Exception):
    """Exception raised when a key sequence string has a syntax error"""

    def __str__(self):
        return ' '.join(self.args)


class KeyAction(object):
    """Class that represents a single 'keyboard' action

    It represents either a PAUSE action (not reallly keyboard) or a keyboard
    action (press or release or both) of a particular key.
    """

    def __init__(self, key, down=True, up=True):
        self.key = key
        if isinstance(self.key, six.string_types):
            self.key = unicode(key)
        self.down = down
        self.up = up

    def _get_key_info(self):
        """Return virtual_key, scan_code, and flags for the action

        This is one of the methods that will be overridden by sub classes"""
        return 0, ord(self.key), KEYEVENTF_UNICODE

    def GetInput(self):
        "Build the INPUT structure for the action"
        actions = 1
        # if both up and down
        if self.up and self.down:
            actions = 2

        inputs = (INPUT * actions)()

        vk, scan, flags = self._get_key_info()

        for inp in inputs:
            inp.type = INPUT_KEYBOARD

            inp._.ki.wVk = vk
            inp._.ki.wScan = scan
            inp._.ki.dwFlags |= flags

        # if we are releasing - then let it up
        if self.up:
            inputs[-1]._.ki.dwFlags |= KEYEVENTF_KEYUP

        return inputs

    def Run(self):
        "Execute the action"
        inputs = self.GetInput()
        return SendInput(
            len(inputs),
            ctypes.byref(inputs),
            ctypes.sizeof(INPUT))

    def _get_down_up_string(self):
        """Return a string that will show whether the string is up or down

        return 'down' if the key is a press only
        return 'up' if the key is up only
        return '' if the key is up & down (as default)
        """
        down_up = ""
        if not (self.down and self.up):
            if self.down:
                down_up = "down"
            elif self.up:
                down_up = "up"
        return down_up

    def key_description(self):
        "Return a description of the key"
        vk, scan, flags = self._get_key_info()
        desc = ''
        if vk:
            if vk in CODE_NAMES:
                desc = CODE_NAMES[vk]
            else:
                desc = "VK %d" % vk
        else:
            desc = "%s" % self.key

        return desc

    def __str__(self):
        parts = []
        parts.append(self.key_description())
        up_down = self._get_down_up_string()
        if up_down:
            parts.append(up_down)

        return "<%s>" % (" ".join(parts))
    __repr__ = __str__


class VirtualKeyAction(KeyAction):
    """Represents a virtual key action e.g. F9 DOWN, etc

    Overrides necessary methods of KeyAction"""

    def _get_key_info(self):
        "Virtual keys have extended flag set"

        # copied more or less verbatim from
        # http://www.pinvoke.net/default.aspx/user32.sendinput
        if ((self.key >= 33 and self.key <= 46) or (self.key >= 91 and self.key <= 93)):
            flags = KEYEVENTF_EXTENDEDKEY
        else:
            flags = 0
        # This works for %{F4} - ALT + F4
        #return self.key, 0, 0

        # this works for Tic Tac Toe i.e. +{RIGHT} SHIFT + RIGHT
        return self.key, MapVirtualKey(self.key, 0), flags


class EscapedKeyAction(KeyAction):
    """Represents an escaped key action e.g. F9 DOWN, etc

    Overrides necessary methods of KeyAction"""

    def _get_key_info(self):
        """EscapedKeyAction doesn't send it as Unicode and the vk and
        scan code are generated differently"""
        vkey_scan = LoByte(VkKeyScan(self.key))

        return (vkey_scan, MapVirtualKey(vkey_scan, 0), 0)

    def key_description(self):
        "Return a description of the key"

        return "KEsc %s" % self.key


class PauseAction(KeyAction):
    "Represents a pause action"

    def __init__(self, how_long):
        self.how_long = how_long

    def Run(self):
        "Pause for the lenght of time specified"
        time.sleep(self.how_long)

    def __str__(self):
        return "<PAUSE %1.2f>" % (self.how_long)
    __repr__ = __str__

    #def GetInput(self):
    #    print `self.key`
    #    keys = KeyAction.GetInput(self)
    #
    #    shift_state = HiByte(VkKeyScan(self.key))
    #
    #    shift_down = shift_state & 0x100  # 1st bit
    #    ctrl_down =  shift_state & 0x80   # 2nd bit
    #    alt_down =  shift_state & 0x40    # 3rd bit
    #
    #    print bin(shift_state), shift_down, ctrl_down, alt_down
    #
    #    print keys
    #    keys = [k for k in keys]
    #
    #    modifiers = []
    #    if shift_down:
    #        keys[0:0] = VirtualKeyAction(VK_SHIFT, up = False).GetInput()
    #        keys.append(VirtualKeyAction(VK_SHIFT, down = False).GetInput())
    #    if ctrl_down:
    #        keys[0:0] = VirtualKeyAction(VK_CONTROL, up = False).GetInput()
    #        keys.append(VirtualKeyAction(VK_CONTROL, down = False).GetInput())
    #    if alt_down:
    #        keys[0:0] = VirtualKeyAction(VK_ALT, up = False).GetInput()
    #        keys.append(VirtualKeyAction(VK_ALT, down = False).GetInput())
    #
    #    print keys
    #    new_keys = (INPUT * len(keys)) ()
    #
    #    for i, k in enumerate(keys):
    #        if hasattr(k, 'type'):
    #            new_keys[i] = k
    #        else:
    #            for sub_key in k:
    #                new_keys[i] = sub_key
    #
    #    return new_keys
    #


def handle_code(code):
    "Handle a key or sequence of keys in braces"

    code_keys = []
    # it is a known code (e.g. {DOWN}, {ENTER}, etc)
    if code in CODES:
        code_keys.append(VirtualKeyAction(CODES[code]))

    # it is an escaped modifier e.g. {%}, {^}, {+}
    elif len(code) == 1:
        code_keys.append(KeyAction(code))

    # it is a repetition or a pause  {DOWN 5}, {PAUSE 1.3}
    elif ' ' in code:
        to_repeat, count = code.rsplit(None, 1)
        if to_repeat == "PAUSE":
            try:
                pause_time = float(count)
            except ValueError:
                raise KeySequenceError('invalid pause time %s' % count)
            code_keys.append(PauseAction(pause_time))

        else:
            try:
                count = int(count)
            except ValueError:
                raise KeySequenceError(
                    'invalid repetition count %s' % count)

            # If the value in to_repeat is a VK e.g. DOWN
            # we need to add the code repeated
            if to_repeat in CODES:
                code_keys.extend(
                    [VirtualKeyAction(CODES[to_repeat])] * count)
            # otherwise parse the keys and we get back a KeyAction
            else:
                to_repeat = parse_keys(to_repeat)
                if isinstance(to_repeat, list):
                    keys = to_repeat * count
                else:
                    keys = [to_repeat] * count
                code_keys.extend(keys)
    else:
        raise RuntimeError("Unknown code: %s" % code)

    return code_keys


def parse_keys(string,
                with_spaces=False,
                with_tabs=False,
                with_newlines=False,
                modifiers=None):
    "Return the parsed keys"

    keys = []
    if not modifiers:
        modifiers = []
    index = 0
    while index < len(string):

        c = string[index]
        index += 1
        # check if one of CTRL, SHIFT, ALT has been pressed
        if c in MODIFIERS.keys():
            modifier = MODIFIERS[c]
            # remember that we are currently modified
            modifiers.append(modifier)
            # hold down the modifier key
            keys.append(VirtualKeyAction(modifier, up=False))
            if DEBUG:
                print("MODS+", modifiers)
            continue

        # Apply modifiers over a bunch of characters (not just one!)
        elif c == "(":
            # find the end of the bracketed text
            end_pos = string.find(")", index)
            if end_pos == -1:
                raise KeySequenceError('`)` not found')
            keys.extend(parse_keys(string[index:end_pos], modifiers=modifiers))
            index = end_pos + 1

        # Escape or named key
        elif c == "{":
            # We start searching from index + 1 to account for the case {}}
            end_pos = string.find("}", index + 1)
            if end_pos == -1:
                raise KeySequenceError('`}` not found')

            code = string[index:end_pos]
            index = end_pos + 1
            keys.extend(handle_code(code))

        # unmatched ")"
        elif c == ')':
            raise KeySequenceError('`)` should be preceeded by `(`')

        # unmatched "}"
        elif c == '}':
            raise KeySequenceError('`}` should be preceeded by `{`')

        # so it is a normal character
        else:
            # don't output white space unless flags to output have been set
            if (c == ' ' and not with_spaces or
                 c == '\t' and not with_tabs or
                 c == '\n' and not with_newlines):
                continue

            # output nuewline
            if c in ('~', '\n'):
                keys.append(VirtualKeyAction(CODES["ENTER"]))

            # safest are the virtual keys - so if our key is a virtual key
            # use a VirtualKeyAction
            #if ord(c) in CODE_NAMES:
            #    keys.append(VirtualKeyAction(ord(c)))

            elif modifiers:
                keys.append(EscapedKeyAction(c))

            else:
                keys.append(KeyAction(c))

        # as we have handled the text - release the modifiers
        while modifiers:
            if DEBUG:
                print("MODS-", modifiers)
            keys.append(VirtualKeyAction(modifiers.pop(), down=False))

    # just in case there were any modifiers left pressed - release them
    while modifiers:
        keys.append(VirtualKeyAction(modifiers.pop(), down=False))

    return keys


def LoByte(val):
    "Return the low byte of the value"
    return val & 0xff


def HiByte(val):
    "Return the high byte of the value"
    return (val & 0xff00) >> 8


def SendKeys(keys,
             pause=0.05,
             with_spaces=False,
             with_tabs=False,
             with_newlines=False,
             turn_off_numlock=True):
    "Parse the keys and type them"
    keys = parse_keys(keys, with_spaces, with_tabs, with_newlines)

    for k in keys:
        k.Run()
        time.sleep(pause)


def main():
    "Send some test strings"

    actions = """
        {LWIN}
        {PAUSE .25}
        r
        {PAUSE .25}
        Notepad.exe{ENTER}
        {PAUSE 1}
        Hello{SPACE}World!
        {PAUSE 1}
        %{F4}
        {PAUSE .25}
        n
        """
    SendKeys(actions, pause=.1)

    keys = parse_keys(actions)
    for k in keys:
        print(k)
        k.Run()
        time.sleep(.1)

    test_strings = [
        "\n"
        "(aa)some text\n",
        "(a)some{ }text\n",
        "(b)some{{}text\n",
        "(c)some{+}text\n",
        "(d)so%me{ab 4}text",
        "(e)so%me{LEFT 4}text",
        "(f)so%me{ENTER 4}text",
        "(g)so%me{^aa 4}text",
        "(h)some +(asdf)text",
        "(i)some %^+(asdf)text",
        "(j)some %^+a text+",
        "(k)some %^+a tex+{&}",
        "(l)some %^+a tex+(dsf)",
        "",
        ]

    for s in test_strings:
        print(repr(s))
        keys = parse_keys(s, with_newlines=True)
        print(keys)

        for k in keys:
            k.Run()
            time.sleep(.1)
        print()

if __name__ == "__main__":

    main()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
from six.moves import zip, map, range  # NOQA
try:
    from collections.abc import MutableSet
except Exception:
    from collections import MutableSet
import weakref
from utool import util_inject
print, rrr, profile = util_inject.inject2(__name__)


class _Link(object):
    __slots__ = ('prev', 'next', 'key', '__weakref__')


class OrderedSet(MutableSet):
    """ Set the remembers the order elements were added
     Big-O running times for all methods are the same as for regular sets.
     The internal self._map dictionary maps keys to links in a doubly linked list.
     The circular doubly linked list starts and ends with a sentinel element.
     The sentinel element never gets deleted (this simplifies the algorithm).
     The prev/next links are weakref proxies (to prevent circular references).
     Individual links are kept alive by the hard reference in self._map.
     Those hard references disappear when a key is deleted from an OrderedSet.

    References:
        http://code.activestate.com/recipes/576696/
        http://code.activestate.com/recipes/576694/
        http://stackoverflow.com/questions/1653970/does-python-have-an-ordered-set
    """

    def __init__(self, iterable=None):
        self._root = root = _Link()  # sentinel node for doubly linked list
        root.prev = root.next = root
        self._map = {}  # key --> link
        if iterable is not None:
            self |= iterable

    def __len__(self):
        return len(self._map)

    def __contains__(self, key):
        return key in self._map

    def add(self, key):
        """ Store new key in a new link at the end of the linked list """
        if key not in self._map:
            self._map[key] = link = _Link()
            root = self._root
            last = root.prev
            link.prev, link.next, link.key = last, root, key
            last.next = root.prev = weakref.proxy(link)

    def append(self, key):
        """ Alias for add """
        return self.add(key)

    def discard(self, key):
        # Remove an existing item using self._map to find the link which is
        # then removed by updating the links in the predecessor and successors.
        if key in self._map:
            link = self._map.pop(key)
            link.prev.next = link.next
            link.next.prev = link.prev

    def __iter__(self):
        # Traverse the linked list in order.
        root = self._root
        curr = root.next
        while curr is not root:
            yield curr.key
            curr = curr.next

    def __reversed__(self):
        # Traverse the linked list in reverse order.
        root = self._root
        curr = root.prev
        while curr is not root:
            yield curr.key
            curr = curr.prev

    def pop(self, last=True):
        if not self:
            raise KeyError('set is empty')
        key = next(reversed(self)) if last else next(iter(self))
        self.discard(key)
        return key

    def __repr__(self):
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, list(self))

    def __eq__(self, other):
        if isinstance(other, OrderedSet):
            return len(self) == len(other) and list(self) == list(other)
        return not self.isdisjoint(other)

    @classmethod
    def union(cls, *sets):
        """
        >>> from utool.util_set import *  # NOQA
        """
        import utool as ut
        lists_ = ut.flatten([list(s) for s in sets])
        return cls(lists_)

    def update(self, other):
        """ union update """
        for item in other:
            self.add(item)

    def __getitem__(self, index):
        """
        Example:
            >>> # ENABLE_DOCTEST
            >>> import utool as ut
            >>> self = ut.oset([1, 2, 3])
            >>> assert self[0] == 1
            >>> assert self[1] == 2
            >>> assert self[2] == 3
            >>> ut.assert_raises(IndexError, self.__getitem__, 3)
            >>> assert self[-1] == 3
            >>> assert self[-2] == 2
            >>> assert self[-3] == 1
            >>> ut.assert_raises(IndexError, self.__getitem__, -4)
        """
        if index < 0:
            iter_ = self.__reversed__
            index_ = -1 - index
        else:
            index_ = index
            iter_ = self.__iter__
        if index_ >= len(self):
            raise IndexError('index %r out of range %r' % (index, len(self)))
        for count, item in zip(range(index_ + 1), iter_()):
            pass
        return item

    def index(self, item):
        """
        Find the index of `item` in the OrderedSet

        Example:
            >>> # ENABLE_DOCTEST
            >>> import utool as ut
            >>> self = ut.oset([1, 2, 3])
            >>> assert self.index(1) == 0
            >>> assert self.index(2) == 1
            >>> assert self.index(3) == 2
            >>> ut.assert_raises(ValueError, self.index, 4)
        """
        for count, other in enumerate(self):
            if item == other:
                return count
        raise ValueError('%r is not in OrderedSet' % (item,))


# alias
oset = OrderedSet



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
try:
    import numpy as np
except ImportError:
    pass
import six
import functools
import sys
from six.moves import builtins
from utool._internal import meta_util_arg
from utool import util_str
from utool import util_inject
print, rrr, profile = util_inject.inject2(__name__)

QUIET        = meta_util_arg.QUIET
VERBOSE      = meta_util_arg.VERBOSE
NO_INDENT    = meta_util_arg.NO_INDENT
SILENT       = meta_util_arg.SILENT


def print_dict(dict_, dict_name=None, **kwargs):
    import utool as ut
    if dict_name is None:
        dict_name = ut.get_varname_from_stack(dict_, N=1)
    dict_repr = util_str.repr4(dict_, **kwargs)
    print(dict_name + ' = ' + dict_repr)

printdict = print_dict
dictprint = print_dict


def print_list(list_, **kwargs):
    import utool as ut
    list_name = ut.get_varname_from_stack(list_, N=1)
    print(list_name + ' = ' + util_str.repr4(list_, **kwargs))


def horiz_print(*args):
    toprint = util_str.horiz_string(args)
    print(toprint)


#def set_indenting_enabled(flag):
#    global NO_INDENT
#    prev_flag = NO_INDENT
#    NO_INDENT = not flag
#    return prev_flag


def _test_indent_print():
    # Indent test code doesnt work in doctest blocks.
    import utool as ut
    flag = ut.ensure_logging()
    print('Checking indent. Should have none')
    with ut.Indenter('[INDENT] '):
        print('Checking indent. Should be indented')
    print('Should no longer be indented')
    text = ut.get_current_log_text()
    # The last line might sometimes be empty or not.
    # Not sure.
    # New hack: had to put in stride. Seems like logs get written
    # with two line breaks now
    last_lines = text.split('\n')[-8::2]
    if last_lines[-1] != '':
        assert False, 'DEV ERROR. REMOVE FIRST LINE INSTEAD OF LAST'
        last_lines = last_lines[:-1]

    #print('last_lines = %r' % (ut.repr3(last_lines)))
    try:
        assert last_lines[0].find('[INDENT] ') == -1, last_lines[0]
        assert last_lines[1].find('[INDENT] ') >= 0, 'did not indent %r' % (last_lines[1],)
        assert last_lines[2].find('[INDENT] ') == -1, last_lines[2]
    except AssertionError:
        print('Error. Last 3 lines')
        print(ut.repr3(last_lines))
        raise
    if not flag:
        ut.stop_logging()


class Indenter(object):
    r"""
    Monkey patches modules injected with print to change the way print behaves.

    Works with utool.inject to allow for prefixing of all within-context
    print statements in a semi-dynamic manner. There seem to be some bugs
    but it works pretty well.

    CommandLine:
        python -m utool.util_print --exec-Indenter

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_print import *  # NOQA
        >>> import utool as ut
        >>> ut.util_print._test_indent_print()
    """
    # THIS IS MUCH BETTER
    def __init__(self, lbl='    ', enabled=True):
        self.enabled = enabled
        if not NO_INDENT or not self.enabled:
            #self.modules = modules
            self.modules = util_inject.get_injected_modules()
            self.old_print_dict = {}
            #self.old_prints_ = {}
            #self.old_printDBG_dict = {}
            self.lbl = lbl
            #self.INDENT_PRINT_ = False

    def start(self):
        # Chain functions together rather than overwriting stdout
        if NO_INDENT or not self.enabled:
            return builtins.print
        def indent_msg(*args):
            mgs = ', '.join(map(six.text_type, args))
            return self.lbl + mgs.replace('\n', '\n' + self.lbl)

        def push_module_functions(dict_, funcname):
            for mod in self.modules:
                try:
                    dict_[mod] = getattr(mod, funcname)
                except KeyError as ex:
                    print('[utool] KeyError: ' + six.text_type(ex))
                    print('[utool] WARNING: module=%r was loaded between indent sessions' % mod)
                except AttributeError as ex:
                    print('[utool] AttributeError: ' + six.text_type(ex))
                    print('[utool] WARNING: module=%r does not have injected utool prints' % mod)

        push_module_functions(self.old_print_dict, 'print')
        for mod in self.old_print_dict.keys():
            # Define the new print function
            @functools.wraps(self.old_print_dict[mod])
            def indent_print(*args):
                self.old_print_dict[mod](indent_msg(', '.join(map(six.text_type, args))))
            setattr(mod, 'print', indent_print)
        return indent_print

        #push_module_functions(self.old_printDBG_dict, 'printDBG')
        #for mod in self.old_printDBG_dict.keys():
        #    @functools.wraps(self.old_printDBG_dict[mod])
        #    def indent_printDBG(msg):
        #        self.old_printDBG_dict[mod](indent_msg(msg))
        #    setattr(mod, 'printDBG', indent_printDBG)

    def stop(self):
        if NO_INDENT or not self.enabled:
            return
        def pop_module_functions(dict_, funcname):
            for mod in six.iterkeys(dict_):
                setattr(mod, funcname, dict_[mod])
        pop_module_functions(self.old_print_dict, 'print')
        #pop_module_functions(self.old_printDBG_dict, 'printDBG')
        #for mod in six.iterkeys(self.old_print_dict):
        #    setattr(mod, 'print', self.old_print_dict[mod])
        #for mod in six.iterkeys(self.old_printDBG_dict):
        #    setattr(mod, 'printDBG', self.old_printDBG_dict[mod])
        return builtins.print

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, type_, value, trace):
        self.stop()
        if trace is not None:
            if VERBOSE:
                print('[util_print] Error in print context manager!: ' + str(value))
            return False  # return a falsey value on error


def printshape(arr_name, locals_):
    arr = locals_[arr_name]
    if isinstance(arr, np.ndarray):
        print(arr_name + '.shape = ' + str(arr.shape))
    else:
        print('len(%s) = %r' % (arr_name, len(arr)))


#class NpPrintOpts(object):
#    def __init__(self, **kwargs):
#        self.orig_opts = np.get_printoptions()
#        self.new_opts = kwargs
#    def __enter__(self):
#        np.set_printoptions(**self.new_opts)
#    def __exit__(self, exc_type, exc_value, exc_traceback):
#        np.set_printoptions(**self.orig_opts)
#        if exc_traceback is not None:
#            print('[util_print] ERROR IN TRACEBACK: ' + str(exc_value))
#            return False


#def full_numpy_repr(arr):
#    with NpPrintOpts(threshold=np.uint64(-1)):
#        arr_repr = repr(arr)
#    return arr_repr


def printVERBOSE(msg, verbarg):
    if VERBOSE or verbarg in sys.argv:
        print(msg)


def printNOTQUIET(msg):
    if not QUIET:
        print(msg)


def printWARN(msg):
    try:
        import colorama
        from colorama import Fore, Style
        colorama.init()
        print(Fore.RED + msg + Style.RESET_ALL)
        colorama.deinit()
    except ImportError:
        print(msg)


def print_filesize(fpath):
    print(util_str.filesize_str(fpath))


def printif(func, condition=VERBOSE and not QUIET):
    """ execute printfunc only if condition=QUIET"""
    if condition:
        print(func())


def print_python_code(text):
    r"""
    SeeAlso:
        print_code
    """
    print_code(text, 'python')


def print_code(text, lexer_name='python'):
    r"""
    Args:
        text (str):

    CommandLine:
        python -m utool.util_print --test-print_python_code

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_print import *  # NOQA
        >>> import utool as ut
        >>> # build test data
        >>> text = ut.read_from(ut.__file__.replace('.pyc', '.py'))
        >>> # execute function
        >>> print_python_code(text)
    """
    print(util_str.highlight_code(text, lexer_name))


def print_difftext(text, other=None):
    """
    Args:
        text (str):

    CommandLine:
        #python -m utool.util_print --test-print_difftext
        #autopep8 ingest_data.py --diff | python -m utool.util_print --test-print_difftext

    """
    if other is not None:
        # hack
        text = util_str.difftext(text, other)
    colortext = util_str.color_diff_text(text)
    try:
        print(colortext)
    except UnicodeEncodeError as ex:  # NOQA
        import unicodedata
        colortext = unicodedata.normalize('NFKD', colortext).encode('ascii', 'ignore')
        print(colortext)


def colorprint(text, color=None):
    r""" provides some color to terminal output

    Args:
        text (str):
        color (str):

    Ignore:
        assert color in ['', 'yellow', 'blink', 'lightgray', 'underline',
        'darkyellow', 'blue', 'darkblue', 'faint', 'fuchsia', 'black', 'white',
        'red', 'brown', 'turquoise', 'bold', 'darkred', 'darkgreen', 'reset',
        'standout', 'darkteal', 'darkgray', 'overline', 'purple', 'green', 'teal',
        'fuscia']

    CommandLine:
        python -c "import pygments.console; print(list(pygments.console.codes.keys()))"
        python -m utool.util_print --exec-colorprint
        python -m utool.util_print --exec-colorprint:1

        import pygments
        print(ut.repr4(list(pygments.formatters.get_all_formatters())))
        print(list(pygments.styles.get_all_styles()))

    Example0:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_print import *  # NOQA
        >>> import pygments.console
        >>> msg_list = list(pygments.console.codes.keys())
        >>> color_list = list(pygments.console.codes.keys())
        >>> [colorprint(text, color) for text, color in zip(msg_list, color_list)]

    Example1:
        >>> # DISABLE_DOCTEST (Windows test)
        >>> from utool.util_print import *  # NOQA
        >>> import pygments.console
        >>> print('line1')
        >>> colorprint('line2', 'red')
        >>> colorprint('line3', 'blue')
        >>> colorprint('line4', 'fuchsia')
        >>> colorprint('line5', 'reset')
        >>> colorprint('line5', 'fuchsia')
        >>> print('line6')
    """
    print(util_str.color_text(text, color))


cprint = colorprint


def print_locals(*args, **kwargs):
    """
    Prints local variables in function.

    If no arguments all locals are printed.

    Variables can be specified directly (variable values passed in) as varargs
    or indirectly (variable names passed in) in kwargs by using keys and a list
    of strings.
    """
    from utool import util_str
    from utool import util_dbg
    from utool import util_dict
    locals_ = util_dbg.get_parent_frame().f_locals
    keys = kwargs.get('keys', None if len(args) == 0 else [])
    to_print = {}
    for arg in args:
        varname = util_dbg.get_varname_from_locals(arg, locals_)
        to_print[varname] = arg
    if keys is not None:
        to_print.update(util_dict.dict_take(locals_, keys))
    if not to_print:
        to_print = locals_
    locals_str = util_str.repr4(to_print)
    print(locals_str)


if __name__ == '__main__':
    """
    CommandLine:
        python -m utool.util_print
        python -m utool.util_print --allexamples
        python -m utool.util_print --allexamples --noface --nosrc """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
import utool
from utool import util_io

from utool.util_io import write_to
write_to2 = utool.write_to


def print_ids():
    print('<')
    print('utool.write_to:         %r' % id(utool.write_to))  # WHY GOOD?
    print('utool.util_io.write_to: %r' % id(utool.util_io.write_to))  # GOOD
    print('util_io.write_to:       %r' % id(util_io.write_to))  # GOOD
    print('write_to:               %r' % id(write_to))   # BAD
    print('write_to2:              %r' % id(write_to2))  # BAD
    print('>')


def print_docstr():
    print('<')
    print('utool.write_to:         %r' % (utool.write_to.__doc__))
    print('utool.util_io.write_to: %r' % (utool.util_io.write_to.__doc__))
    print('util_io.write_to:       %r' % (util_io.write_to.__doc__))
    print('write_to:               %r' % (write_to.__doc__))
    print('write_to2:              %r' % (write_to2.__doc__))
    print('>')


def reloading_test1():
    print('TEST1')
    print_ids()
    utool.rrrr()
    print_ids()


def reloading_test2():
    print('TEST2')
    print_ids()
    import imp
    # Causes utool.write_to to be bad
    # because utool doesnt do a rrr itself
    imp.reload(utool.util_io)
    print_ids()


def docstr_test1():
    print_ids()
    utool.rrr()
    print_ids()


def docstr_test2():
    print_ids()
    utool.rrr()
    print_ids()


if __name__ == '__main__':
    utool.print_object_size(utool)
    reloading_test1()
    #reloading_test2()
    reloading_test1()
    reloading_test1()
    #reloading_test2()
    utool.print_object_size(utool)
    pass
